%
% Copyright (C) 2017 P.Bernal-Polo
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%

% 
% File:   OE_Tester.m
% Author: P.Bernal-Polo
% 
% Created on February 19, 2017, 2:05 PM
%


classdef OE_Tester
  % OE_Tester: class used for testing the performance of an OrientationEstimator algorithm
  %   
  %   This class implements some methods with the objective of testing, in
  %   an easy way, the performance of an algorithm of the
  %   OrientationEstimator class
  
  properties( Access = private )
    title = 'simulation';  % title used to save results
    f = [1 10 100 1000];  % vector of frequency samples (Hz)
    convergenceThreshold = 3*pi/180.0;  % threshold for which convergence is considered achieved (it is stored in radians, but introduced in degrees)
    maxConvergenceUpdates = 10000;  % maximum number of updates when trying to achieve convergence
    Tsim = 1;  % simulation time (s)
    dtdtsim = 10;  % (dt/dtsim) measure of how big is the update time step (dt) vs the simulation time step (dtsim)
    Ntimes = 100;  % times to repeat the simulation for each frequency
    QwMax = 1e0;  % maximum angular velocity variance per unit of time (rad^2/s^3)
    QvMax = 1e-2;  % maximum variance of the measured vector
    Rw = 1e-4;  % variance in angular velocity measurement (rad^2/s^2)
    Rv = 1e-4;  % variance in the vector measurement
  end
  
  
  
  methods
    
    % Function: set_title
    % sets the title for files to save results
    % inputs:
    %  title: title of the file to save results
    % outputs:
    %  obj: OE_Tester object with the new title
    function obj = set_title( obj , title )
      if ~ischar(title)
        error('Error in "set_title()": title must be a string');
      end
      obj.title = title;
    end
    
    
    % Function: set_f
    % sets the frequencies for which the performance of the filter will be evaluated.
    % the frequency vector introduced here will be modified choosing the
    % nearest frequencies that produce an update at obj.Tsim
    % inputs:
    %  fin: vector of frequencies (Hz)
    % outputs:
    %  obj: OE_Tester object with the new f
    function obj = set_f( obj , fin )
      if size( fin , 1 ) > 1
        error('Error in "set_f()": please, provide the frequency vector in a row');
      end
      if sum( fin < 1.0/obj.Tsim )
        error('Error in "set_f()": all frequencies must be greater than 1.0/obj.Tsim');
      end
      obj.f = fin;
    end
    
    
    % Function: set_convergenceThreshold
    % sets the convergence threshold for the convergence step of the simulation
    % inputs:
    %  convergenceThreshold: degrees difference for which the convergence is considered achieved (degrees)
    % outputs:
    %  obj: OE_Tester object with the new convergenceThreshold
    function obj = set_convergenceThreshold( obj , convergenceThreshold )
      if convergenceThreshold < 0
        error('Error in "set_convergenceThreshold()": convergenceThreshold must be positive.');
      end
      obj.convergenceThreshold = convergenceThreshold*pi/180.0;
    end
    
    
    % Function: set_maxConvergenceUpdates
    % sets the maximum number of updates for the convergence step of the simulation
    % inputs:
    %  maxConvergenceUpdates: maximum number of updates for the convergence step
    % outputs:
    %  obj: OE_Tester object with the new maxConvergenceUpdates
    function obj = set_maxConvergenceUpdates( obj , maxConvergenceUpdates )
      if maxConvergenceUpdates < 0
        error('Error in "set_maxConvergenceUpdates()": maxConvergenceUpdates must be positive.');
      end
      obj.maxConvergenceUpdates = maxConvergenceUpdates;
    end
    
    
    % Function: set_Tsim
    % sets the time for which the simulation is run
    % inputs:
    %  Tsim: period of time for which the simulation is run (s)
    % outputs:
    %  obj: OE_Tester object with the new Tsim
    function obj = set_Tsim( obj , Tsim )
      if Tsim < 1.0/min( obj.f )
        error('Error in "set_Tsim()": Tsim cannot be less than the minimum frequency.');
      end
      if Tsim <= 0.0
        error('Error in "set_Tsim()": Tsim must be greater than 0.0');
      end
      obj.Tsim = Tsim;
    end
    
    
    % Function: set_simStepsPerUpdate
    % sets the simulation steps carried before each OrientationEstimator update
    % inputs:
    %  dtdtsimin: number of simulation steps carried before each update (that turns out to be equal to dt/dtsim)
    % outputs:
    %  obj: OE_Tester object with the new dtdtsim
    function obj = set_simStepsPerUpdate( obj , dtdtsimin )
      if ceil(dtdtsimin) ~= floor(dtdtsimin)  ||  dtdtsimin <= 0
        error('Error in "set_dtdtsim": dtdtsim must be a positive integer');
      end
      obj.dtdtsim = round( dtdtsimin );
    end
    
    
    % Function: set_Ntimes
    % sets the number of trajectories generated for each frequency in order of testing the OrientationEstimator
    % inputs:
    %  Ntimes: number of trajectories generated for each frequency
    % outputs:
    %  obj: OE_Tester object with the new Ntimes
    function obj = set_Ntimes( obj , Ntimes )
      if Ntimes < 1
        error('Error in "set_Ntimes": Ntimes must be 1 or greater.');
      end
      obj.Ntimes = round( Ntimes );
    end
    
    
    % Function: set_QwMax
    % sets the maximum variance per time unit for the process noise in the angular velocity
    % inputs:
    %  QwMax: maximum variance per time unit for the process noise in the angular velocity per unit of time (rad^2/s^3)
    % outputs:
    %  obj: OE_Tester object with the new QwMax
    function obj = set_QwMax( obj , QwMax )
      if QwMax < 0
        error('Error in "set_QwMax()": QwMax must be greater than 0.0');
      end
      obj.QwMax = QwMax;
    end
    
    
    % Function: set_QvMax
    % sets the maximum variance for the process noise in the vector
    % inputs:
    %  QvMax: maximum variance for the process noise in the vector
    % outputs:
    %  obj: OE_Tester object with the new QvMax
    function obj = set_QvMax( obj , QvMax )
      if QvMax < 0
        error('Error in "set_QvMax()": QvMax must be greater than 0.0');
      end
      obj.QvMax = QvMax;
    end
    
    
    % Function: set_Rw
    % sets the variance for the angular velocity measurement
    % inputs:
    %  Rw: variance of the angular velocity measurement (rad^2/s^2)
    % outputs:
    %  obj: OE_Tester object with the new Rw
    function obj = set_Rw( obj , Rw )
      if Rw < 0
        error('Error in "set_Rw()": Rw must be greater than 0.0');
      end
      obj.Rw = Rw;
    end
    
    
    % Function: set_Rv
    % sets the variance for the vector measurement
    % inputs:
    %  Rv: variance for the vector measurement
    % outputs:
    %  obj: OE_Tester object with the new Rv
    function obj = set_Rv( obj , Rv )
      if Rv < 0
        error('Error in "set_Rv()": Rv must be greater than 0.0');
      end
      obj.Rv = Rv;
    end
    
    
    % Function: test
    % uses the OE_Tester parameters to test an OrientationEstimator, and
    % saves the results in a file
    % inputs:
    %  estimator: the OrientationEstimator object for which we want to find its performance
    %  infoUpdateTime: time step to show information about the simulation progress
    function test( obj , myEstimator , infoUpdateTime )
      if ~isnumeric(infoUpdateTime)
        error('Error in "test()": infoUpdateTime must be a number.');
      end
      % first of all, let me write about the decisions taken for this
      % simulation.
      % The first thought was to generate a sequence of states that
      % generate a smooth and continuous trajectory for the orientations.
      % Then we would take simulated measurements, and use these sequences
      % for updating our estimator. There are two major problems with this
      % approach: (keep reading)
      % - we can not assert the coincidence of the final step of the
      % simulation with the final update
      % - we can not assert the coincidence of all the estimation updates
      % for all the frequencies, with a given time step of the sequence of
      % simulated states
      % A second approach to overcome these problems is to generate a
      % different trajectory for each frequency. It is done defining an
      % integer quantity dtdtsim, that determines the simulation steps
      % performed before each estimator update (asserting the equality
      % "n*dtsim = dt", with n an integer).
      % Then, the frequencies are transformed to satisfy the equality
      % "m*dt = Tsim", with m an integer.
      % After explaining this, we can start with the code
      
      % we require the condition that the final update happens at the end of the simulation:
      %    n*dt = Tsim  =>  f = 1.0/dt = n/Tsim = round(Tsim/dt)/Tsim = round(Tsim*f)/Tsim
      fs = round(obj.Tsim*obj.f)/obj.Tsim;
      % we do not want repeated frequencies
      fs = unique( fs );
      % and now we can compute the number of frequencies
      Nfrequencies = size( fs , 2 );
      
      % we compute the sigmas of the measurement covariances
      rw = sqrt( obj.Rw );
      rv = sqrt( obj.Rv );
      
      % we will save our error definitions in these vectors
      NnoConvergence = zeros( Nfrequencies ,1);
      Ncrashes = zeros( Nfrequencies , 1 );
      TConv = zeros( Nfrequencies ,1);
      TConv2 = zeros( Nfrequencies ,1);
      errorQ = zeros( Nfrequencies ,1);
      errorQ2 = zeros( Nfrequencies ,1);
      
      % we will use this variable to display some information
      msg = '';
      
      % we take the initial time to give time estimations
      timer = tic;
      beginingTime = toc(timer);
      lastTime = beginingTime;
      
      % we do it several times to get a good performance estimation
      for n=1:obj.Ntimes
        
        % we test for each frequency
        for nf=1:Nfrequencies
          % we compute the update time step
          dt = 1.0/fs(nf);
          
          % we generate random sigmas for the process using the maximum variances
          nw = rand()*obj.QwMax;
          nv = rand()*obj.QvMax;
          
          % we take a random initial orientation, and zero angular velocity
          q0 = normrnd( 0 , 1 , [4 1]);
          q0 = q0/norm(q0);
          w0 = [0; 0; 0];
          
          % and we initialize the estimator
          RwIn = eye(3,3)*obj.Rw;
          RvIn = eye(3,3)*obj.Rv;
          myEstimator = myEstimator.initialize( RwIn , RvIn );
          
          % the first simulation part is to maintain the sensor static until
          % it reaches convergence or too many updates
          theta = 1.0e2;
          Nupdates = 0;
          crashed = false;
          while theta > obj.convergenceThreshold  &&  Nupdates < obj.maxConvergenceUpdates
            % first we generate a random unit vector measured in the
            % external reference frame (it could also have a norm different from one, but we do so for simplicity)
            v = normrnd( 0.0 , 1.0 , [3 1] );
            v = v/norm(v);
            % we compute the simulated measurement
            [ vm , wm ] = OE_Tester.get_measurement( q0 , w0 , v , nv , rv , rw );
            % and we update the filter with the measurements
            myEstimator = myEstimator.update( v , vm , wm , dt );
            Nupdates = Nupdates + 1;
            % we take the estimated orientation, and we check if the algorithm has crashed
            if OE_Tester.hasCrashed( myEstimator.get_q() )
              crashed = true;
              break;
            end
            % we compute the error in the orientation estimation
            theta = OE_Tester.thetaQ( q0 , myEstimator.get_q() );
          end
          % we check if there has been convergence
          if crashed || Nupdates >= obj.maxConvergenceUpdates
            % if not, we add one to the no-convergences, and we go with the next frequency
            NnoConvergence(nf) = NnoConvergence(nf) + 1;
            continue;
          end
          
          % we go ahead with the metrics computations, and with the simulation
          TConv(nf) = TConv(nf) + Nupdates;
          TConv2(nf) = TConv2(nf) + Nupdates*Nupdates;
          
          % we take the initial state
          q = q0;
          w = w0;
          
          % and the second step of the simulation is to generate a random
          % orientation trajectory, and try to estimate it
          %  we need the simulation time step
          dtsim = dt/obj.dtdtsim;
          %  and the number of updates
          Nupdates = round( obj.Tsim*fs(nf) );
          %  we also need the integral of thetaQ, and thetaQ^2
          thQ = 0.0;
          thQ2 = 0.0;
          %  now we can start iterating
          for nup=1:Nupdates
            % first we iterate in the simulation (trajectory generation)
            for nt=1:obj.dtdtsim
              [ q , w ] = OE_Tester.trajectoryStep( q , w , dtsim , nw );
            end
            % we generate a random unit vector measured in the external reference frame
            v = normrnd( 0.0 , 1.0 , [3 1] );
            v = v/norm(v);
            % then we simulate the measurement
            [ vm , wm ] = OE_Tester.get_measurement( q , w , v , nv , rv , rw );
            % finally we update the estimator with the measurement
            myEstimator = myEstimator.update( v , vm , wm , dt );
            % we take the estimated orientation, and we check if the algorithm has crashed
            if OE_Tester.hasCrashed( myEstimator.get_q() )
              crashed = true;
              break;
            end
            % and we add the errors
            %   angle between orientations
            theta = OE_Tester.thetaQ( q , myEstimator.get_q() );
            thQ = thQ + theta;
            thQ2 = thQ2 + theta*theta;
            % we repeat this for Nupdates
          end
          % we check if there has been a crash
          if crashed
            Ncrashes(nf) = Ncrashes(nf) + 1;
            continue;
          end
          
          % now we can compute the error definitions
          dtTsim = dt/obj.Tsim;
          %   error in orientation
          errorQ(nf) = errorQ(nf) + thQ*dtTsim;
          errorQ2(nf) = errorQ2(nf) + thQ2*dtTsim;
          
        end
        
        % if enough time has passed
        if infoUpdateTime > 0.0
          currentTime = toc(timer);
          if  currentTime-lastTime > infoUpdateTime
            % we display some information about resting time, and about the simulation progress
            eTime2end = (currentTime-beginingTime)*(obj.Ntimes-n)/n;
            eTime2endD = floor( eTime2end/86400.0 );
            eTime2endH = floor( ( eTime2end-eTime2endD*86400.0 )/3600.0 );
            eTime2endM = floor( ( eTime2end-eTime2endD*86400.0-eTime2endH*3600.0 )/60.0 );
            eTime2endS = ceil( eTime2end-eTime2endD*86400.0-eTime2endH*3600.0-eTime2endM*60.0 );
            proportion = n/obj.Ntimes;
            msg0 = [ '[' repmat( '-' , 1 , round(proportion*50.0)-1 ) '>' blanks( 50-round(proportion*50.0) ) sprintf( '] (%f) ' , proportion ) 'Estimated remaining time:' ];
            if eTime2endD > 0
              msg0 = [ msg0  sprintf( ' %d d' , eTime2endD ) ];
            end
            if eTime2endH > 0
              msg0 = [ msg0  sprintf( ' %d h' , eTime2endH ) ];
            end
            if eTime2endM > 0
              msg0 = [ msg0  sprintf( ' %d m' , eTime2endM ) ];
            end
            msg0 = [ msg0  sprintf( ' %d s\n' , eTime2endS ) ];
            fprintf( repmat( '\b' , 1 , length(msg) ) );
            fprintf( msg0 );
            msg = msg0;
            
            % we update lastTime for the next infoUpdate
            lastTime = currentTime;
          end
        end
        
      end
      
      % we save the results, but first we correct the measurements
      for nf=1:Nfrequencies
        TConv(nf) = TConv(nf)/fs(nf);
        TConv2(nf) = TConv2(nf)/( fs(nf)*fs(nf) );
        errorQ(nf) = errorQ(nf)*180.0/pi;
        errorQ2(nf) = errorQ2(nf)*180.0/pi*180.0/pi;
      end
      obj.save_results( NnoConvergence , Ncrashes , TConv , TConv2 , errorQ , errorQ2 );
      
      % we display the time taken
      if infoUpdateTime > 0.0
        totalTime = toc(timer);
        timeD = floor( totalTime/86400 );
        timeH = floor( (totalTime-timeD*86400)/3600 );
        timeM = floor( (totalTime-timeD*86400-timeH*3600)/60 );
        timeS = ceil( totalTime-timeD*86400-timeH*3600-timeM*60 );
        fprintf( repmat('\b', 1 , length(msg) ) );
        if timeD
          fprintf( 'Time taken: %d d %d h %d m %d s\n' , [ timeD  timeH  timeM  timeS ] );
        elseif timeH
          fprintf( 'Time taken: %d h %d m %d s\n' , [ timeH  timeM  timeS ] );
        elseif timeM
          fprintf( 'Time taken: %d m %d s\n' , [ timeM  timeS ] );
        elseif timeS
          fprintf( 'Time taken: %d s\n' , timeS );
        end
      end
      
    end
    
    
    % Function: save_results
    % used to save results of the test method
    % inputs:
    %  NnoConv: vector (for each frequency) with number of non-achieved convergences at the second step of the simulation
    %  Ncrashes: vector (for each frequency) with number of algorithm crashes
    %  eTs: vector (for each frequency) with the sum of convergence times
    %  eT2s: vector (for each frequency) with the sum of squared convergence times
    %  eQs: vector (for each frequency) with the sum of errors in orientation estimations
    %  eQ2s: vector (for each frequency) with the sum of squared errors in orientation estimations
    function save_results( obj , NnoConv , Ncrashes , eTs , eT2s , eQs , eQ2s )
      % we open the file
      fileID = fopen( strcat(obj.title,'.dat') , 'w' );
      
      % we print the information about the simulation
      fprintf( fileID , '%%title = %s;\n' , obj.title );
      fprintf( fileID , '%%convergenceThreshold = %e;\n' , obj.convergenceThreshold*180.0/pi );
      fprintf( fileID , '%%maxConvergenceUpdates = %d;\n' , obj.maxConvergenceUpdates );
      fprintf( fileID , '%%Tsim = %e;\n' , obj.Tsim );
      fprintf( fileID , '%%dtdtsim = %d;\n' , obj.dtdtsim );
      fprintf( fileID , '%%Ntimes = %d;\n' , obj.Ntimes );
      fprintf( fileID , '%%QwMax = %e;\n' , obj.QwMax );
      fprintf( fileID , '%%QvMax = %e;\n' , obj.QvMax );
      fprintf( fileID , '%%Rw = %e;\n' , obj.Rw );
      fprintf( fileID , '%%Rv = %e;\n' , obj.Rv );
      
      % now, for each frequency we save:
      fprintf( fileID , '%% frequency (Hz) | n no convergences | n crashes | mean convergence time (s) | sigma convergence time (s) | sigma mean convergence time (s) | mean error in orientation (degrees) | sigma of error in orientation (degrees) | sigma mean error in orientation (degrees)\n' );
      for nf=1:length(obj.f)
        NsamplesT = obj.Ntimes-NnoConv(nf);
        if NsamplesT == 0
          NsamplesT = 1;
        end
        NsamplesSim = obj.Ntimes-Ncrashes(nf);
        if NsamplesSim == 0
          NsamplesSim = 1;
        end
        % - frequency (Hz)
        fprintf( fileID , ' %e' , obj.f(nf) );
        % - number of no convergences
        fprintf( fileID , ' %d' , NnoConv(nf) );
        % - number of crashes
        fprintf( fileID , ' %d' , Ncrashes(nf) );
        % - mean of the convergence time (s)
        meT = eTs(nf)/NsamplesT;
        fprintf( fileID , ' %e' , meT );
        % - sigma of the convergence time (s)
        seT = sqrt( eT2s(nf)/NsamplesT - meT*meT );
        fprintf( fileID , ' %e' , seT );
        % - sigma of the computation of the mean convergence time (s)
        smeT = seT/sqrt(NsamplesT);
        fprintf( fileID , ' %e' , smeT );
        % - mean of the error in orientation estimation (degrees)
        meQ = eQs(nf)/NsamplesSim;
        fprintf( fileID , ' %e' , meQ );
        % - sigma of the computation of the mean error in orientation estimation (degrees)
        seQ = sqrt( eQ2s(nf)/NsamplesSim - meQ*meQ );
        fprintf( fileID , ' %e' , seQ );
        % - sigma of the computation of the mean error in orientation estimation (degrees)
        smeQ = seQ/sqrt(NsamplesSim);
        fprintf( fileID , ' %e' , smeQ );
        fprintf( fileID , '\n' );
      end
      
      % and finally, we close the file
      fclose(fileID);
      
    end
    
    
    % Function: multithread_test
    % this method performs the test function in multiple threads (see the test method)
    % inputs: 
    %  estimator: the OrientationEstimator object for which we want to find its performance
    %  infoUpdateTime: time step to show information about the simulation progress
    %  Nthreads: number of threads used to process the test method
    function multithread_test( obj , estimator , infoUpdateTime , Nthreads )
      
      if Nthreads > 0
        % we create an array of OE_Tester's, and estimators
        for i=1:Nthreads
          tester( i ) = obj;
          threadEstimator( i ) = estimator;
        end
        threadTimes = floor( obj.Ntimes/Nthreads );
        % but each one have different title and Ntimes properties
        totalTimes = 0;
        for i=2:Nthreads
          tester(i) = tester(i).set_title( [ obj.title  sprintf('%d',i) ] );
          tester(i) = tester(i).set_Ntimes( threadTimes );
          totalTimes = totalTimes + threadTimes;
        end
        tester(1) = tester(1).set_title( [ obj.title  '1' ] );
        tester(1) = tester(1).set_Ntimes( obj.Ntimes-totalTimes );
        
        % we execute the test in parallel
        parfor i=1:Nthreads
          if i==1
            tester(i).test( threadEstimator(i) , infoUpdateTime );
          else
            tester(i).test( threadEstimator(i) , -1 );
          end
        end
        
        % after completing the tests, we combine the information
        %   we are interested in the next values
        fs = zeros( 1 , length(obj.f) );
        NnoConv = zeros( length(obj.f) , 1 );
        Ncrashes = zeros( length(obj.f) , 1 );
        eTs = zeros( length(obj.f) , 1 );
        eT2s = zeros( length(obj.f) , 1 );
        eQs = zeros( length(obj.f) , 1 );
        eQ2s = zeros( length(obj.f) , 1 );
        %   for each file created by each thread, we read the corresponding values
        for i=1:Nthreads
          % we open the file
          fileID = fopen( strcat( tester(i).title , '.dat' ) , 'r' );
          % we skip lines starting by '%'
          line = '%';
          while line(1) == '%'
            line = fgets( fileID );
          end
          % for each line (frequency)
          for nf=1:length(obj.f)
            % we read the data
            row = str2num( line );
            % and we add it to the global data
            NsamplesT = tester(i).Ntimes-row(2);
            if NsamplesT == 0
              NsamplesT = 1;
            end
            NsamplesSim = tester(i).Ntimes-row(3);
            if NsamplesSim == 0
              NsamplesSim = 1;
            end
            fs(nf) = row(1);
            NnoConv(nf) = NnoConv(nf) + row(2);
            Ncrashes(nf) = Ncrashes(nf) + row(3);
            eTs(nf) = eTs(nf) + row(4)*NsamplesT;
            eT2s(nf) = eT2s(nf) + (row(5)*row(5)+row(4)*row(4))*NsamplesT;
            eQs(nf) = eQs(nf) + row(7)*NsamplesSim;
            eQ2s(nf) = eQ2s(nf) + (row(8)*row(8)+row(7)*row(7))*NsamplesSim;
            % and we get the new line
            line = fgets( fileID );
          end
          % we close the file
          fclose( fileID );
          % and we delete the file
          delete( strcat( tester(i).title , '.dat' ) );
        end
        % finally we store the combined values
        obj = obj.set_f( fs );
        obj.save_results( NnoConv , Ncrashes , eTs , eT2s , eQs , eQ2s );
        
      else
        obj.test_update( estimator , infoUpdateTime );
      end
      
    end
    
    
  end
  
  
  
  methods( Static )
    
    % Function: trajectoryStep
    % generates a trajectory step from a previous known state
    % inputs:
    %  q: quaternion describing the real orientation
    %  w: angular velocity measured in the sensor reference frame (rad/s)
    %  dtsim: time step of the simulation (s)
    %  nw: standard deviation of the process noise for the angular velocity (rad*s^(-3/2))
    % outputs:
    %  q: new quaternion describing the real orientation of the system (it satisfies qnew = q*qw)
    %  w: new angular velocity (rad/s)
    function [ q , w ] = trajectoryStep( q , w , dtsim , nw )
      % we compute the next angular velocity using the previous one, and the process noise
      % it is necessary to multiply by the square root to get the variance to increase linearly over time
      % (the variance after a time T does not change for different dtsim choices)
      w = w + normrnd( 0 , nw , [3 1] )*sqrt(dtsim);  
      
      % we compute the next orientation using the previous orientation, and the angular velocity
      %   w norm computation
      wnorm = norm( w );
      %   we compute qw
      if wnorm ~= 0.0
        wdt05 = 0.5*wnorm*dtsim;
        swdt = sin(wdt05)/wnorm;
        qw = [ w*swdt; cos(wdt05) ];
      else
        qw = [0; 0; 0; 1];
      end
      %   we compute the new state (q*qw,w)
      q = [q(4)*qw(1:3) + qw(4)*q(1:3) + cross( q(1:3) , qw(1:3) );
           q(4)*qw(4) - q(1:3)'*qw(1:3)];
      q = q/norm(q);  % we make sure that the quaternion is normalized
    end
    
    
    % Function: get_measurement
    % generates measurements from a known state
    % inputs:
    %  q: quaternion describing the orientation of the system (transform vectors from the sensor reference frame to the external reference frame)
    %  w: real angular velocity measured in the sensor reference frame (rad/s)
    %  v: real measured vector expressed in the external reference frame
    %  nv: standard deviation of the process noise for the measured vector
    %  rv: standard deviation of the vector measurement
    %  rw: standard deviation of the gyroscope measurements (rad/s)
    % outputs:
    %  vm: simulated vector measurement
    %  wm: simulated gyroscope measurement (rad/s)
    function [ vm , wm ] = get_measurement( q , w , v , nv , rv , rw )
      % we compute the transposed rotation matrix
      RT = [-q(2)*q(2)-q(3)*q(3)   q(1)*q(2)+q(3)*q(4)   q(1)*q(3)-q(2)*q(4);
             q(1)*q(2)-q(3)*q(4)  -q(1)*q(1)-q(3)*q(3)   q(2)*q(3)+q(1)*q(4);
             q(1)*q(3)+q(2)*q(4)   q(2)*q(3)-q(1)*q(4)  -q(1)*q(1)-q(2)*q(2)];
      RT = RT + RT + eye(3);
      % we generate an additional vector acting on the system
      qv = normrnd( 0 , nv , [3 1] );
      % we obtain the simulated measurements
      vm = RT*( qv + v ) + normrnd( 0 , rv , [3 1] );
      wm = w + normrnd( 0 , rw , [3 1] );
    end
    
    
    % Function: hasCrashed
    % checks if there has been a crash in the algorithm
    % inputs:
    %  q: quaternion computed with the algorithm
    % outputs:
    %  crashed: boolean variable. It will be true if it has crashed; false if not
    function crashed = hasCrashed( q )
      % we will consider a crash in the algorithm if:
      % - q is not of unit norm
      % - q is a NaN
      crashed = (norm(q) > 2.0  ||  ~isreal(q));
    end
    
    
    % Function: thetaQ
    % computes the error in the overall orientation estimation
    % inputs:
    %  qr: current real quaternion describing the orientation in the simulation
    %  qe: current estimated quaternion describing the orientation in the simulation
    % (these quaternions transform vectors from the sensor reference frame to the external reference frame)
    % outputs:
    %  theta: angle between real and estimated orientations (rad)
    function theta = thetaQ( qr , qe )
      % compute the angle between the real orientation and the estimated one (rad)
      delta0re = qr'*qe;
      if delta0re < 0  % the next computation only makes sense for a positive delta0re
        delta0re = -delta0re;
      end
      if delta0re > 1.0
        delta0re = 1.0;
      end
      theta = 2.0*acos(delta0re);
    end
    
    
  end
  
  
  
end






